#ifndef	__HOSTENV__
#define	__HOSTENV__

#ifndef	__
# ifndef __STDC__
#  define __(x) ()
#  ifdef __GNUC__
#   define const    __const
#   define volatile __volatile
#  else
#   define const
#   define volatile
#  endif
# else
#  define __(x) x
# endif
#endif

#include "libsupport.h"
#include "sysprotos.h"

#endif	/* !__HOSTENV__ */
