

z# exit 0

****  Purify'd /home/oj287/mea/zmailer/router/router  (pid 3544)  ****
Purify: Searching for all memory leaks...

There are 20718 leaked bytes (0.0664% of the 31210476 allocated bytes in the heap)

      16 bytes (1222 times).     
      Last memory leak at 0x1fdba58 
Purify (mlk): 19552 total bytes lost, allocated from:
	malloc       [router, pc=0x1d40c]
	emalloc      [line 23, emalloc.c, pc=0xbaf14]
	tmalloc      [line 202, allocate.c, pc=0xb9b38]
	s_copy_tree  [line 86, listtrees.c, pc=0x974ac]
	s_copy_tree  [line 99, listtrees.c, pc=0x97780]
	assign       [line 587, interpret.c, pc=0x85abc]

      1 byte (1092 times).     Last memory leak at 0x145e7a0 
Purify (mlk): 1092 total bytes lost, allocated from:
	malloc       [router, pc=0x1d40c]
	emalloc      [line 23, emalloc.c, pc=0xbaf14]
	strsave      [line 365, allocate.c, pc=0xbaa28]
	s_copy_tree  [line 90, listtrees.c, pc=0x975e4]
	s_copy_tree  [line 99, listtrees.c, pc=0x97780]
	assign       [line 587, interpret.c, pc=0x85abc]

      13 bytes (2 times).     Last memory leak at 0x15de28 
Purify (mlk): 26 total bytes lost, allocated from:
	malloc       [router, pc=0x1d40c]
	calloc       [router, pc=0x1ddbc]
	_tzload      [time_comm_r.c, pc=0xef68617c]
	_ltzset_r    [libc.so.1, pc=0xef6840d0]
	localtime_r  [libc.so.1, pc=0xef682fa0]
	ctime_r      [libc.so.1, pc=0xef681eb8]

      8 bytes (3 times).     Last memory leak at 0x132560 
Purify (mlk): 24 total bytes lost, allocated from:
	malloc       [router, pc=0x1d40c]
	tzcpy        [time_comm_r.c, pc=0xef6846e4]
	_ltzset_r    [libc.so.1, pc=0xef6842d8]
	localtime_r  [libc.so.1, pc=0xef682fa0]
	ctime_r      [libc.so.1, pc=0xef681eb8]
	vsyslog      [libc.so.1, pc=0xef664c38]

      8 bytes (3 times).     Last memory leak at 0x799bd8 
Purify (mlk): 24 total bytes lost, allocated from:
	malloc       [router, pc=0x1d40c]
	tzcpy        [time_comm_r.c, pc=0xef6846e4]
	_ltzset_r    [libc.so.1, pc=0xef68413c]
	localtime_r  [libc.so.1, pc=0xef682fa0]
	ctime_r      [libc.so.1, pc=0xef681eb8]
	vsyslog      [libc.so.1, pc=0xef664c38]

There are 33984 potentially leaked bytes (0.109% of the 31210476 allocated bytes in the heap)

      16 bytes (66 times).     
      Last memory leak at 0x1f9c768 
Purify (plk): 1056 total bytes potentially lost, allocated from:
	malloc       [router, pc=0x1d40c]
	emalloc      [line 23, emalloc.c, pc=0xbaf14]
	tmalloc      [line 202, allocate.c, pc=0xb9b38]
	s_copy_tree  [line 86, listtrees.c, pc=0x974ac]
	s_copy_tree  [line 99, listtrees.c, pc=0x97780]
	assign       [line 587, interpret.c, pc=0x85abc]

Purify (plk): 128 bytes at 0x163df0 potentially lost, allocated from:
	malloc       [router, pc=0x1d40c]
	calloc       [router, pc=0x1ddbc]
	NisSharedCache::NisSharedCache(char*,nis_error*) [libnsl.so.1, pc=0xef775d74]
	__nis_CacheInit [libnsl.so.1, pc=0xef778bb0]
	__nis_CacheBind [libnsl.so.1, pc=0xef778cc0]
	__nis_make_binding [libnsl.so.1, pc=0xef75f090]

Purify [plk]: 32800 bytes potentially lost,
       allocated by malloc in 4 chunks
       whose messages were suppressed with a .purify file.
Purify Heap Analysis (combining suppressed and unsuppressed chunks)

                         Chunks      Bytes
              Leaked       2322      20718
  Potentially Leaked         71      33984
              In-Use       5108   31155774
  ----------------------------------------
     Total Allocated       7501   31210476
 
****  Purify'd /home/oj287/mea/zmailer/router/router  (pid 3544)  ****
  * No access errors.
  * Basic memory usage:
      869132 code
      208036 data/bss
    32808960 heap
        5376 stack
  * Shared library memory usage:
      803596 router	(shared code)
      142500 router	(private data)
      534730 libnsl_pure_2158_53.so.1	(shared code)
       47664 libnsl_pure_2158_53.so.1	(private data)
       49747 libresolv_pure_2158_53.so	(shared code)
        6780 libresolv_pure_2158_53.so	(private data)
       77695 libsocket_pure_2158_53.so.1	(shared code)
        4400 libsocket_pure_2158_53.so.1	(private data)
      593127 libc_pure_2158_53.so.1	(shared code)
       36112 libc_pure_2158_53.so.1	(private data)
        1172 libdl_pure_2158_53.so.1	(shared code)
           0 libdl_pure_2158_53.so.1	(private data)
        5157 libinternal_stubs.so	(shared code)
         272 libinternal_stubs.so	(private data)
        9897 libintl_pure_2158_53.so.1	(shared code)
        2564 libintl_pure_2158_53.so.1	(private data)
       54831 libucb_pure_2158_53.so.1	(shared code)
        3816 libucb_pure_2158_53.so.1	(private data)
       77640 libelf_pure_2158_53.so.1	(shared code)
        3832 libelf_pure_2158_53.so.1	(private data)
       28161 libw_pure_2158_53.so.1	(shared code)
        1124 libw_pure_2158_53.so.1	(private data)
       11750 nss_files_pure_2158_53.so.1	(shared code)
         852 nss_files_pure_2158_53.so.1	(private data)
       27977 nss_nisplus_pure_2158_53.so.1	(shared code)
        1360 nss_nisplus_pure_2158_53.so.1	(private data)
       14113 switch_pure_2158_53.so	(shared code)
        2006 switch_pure_2158_53.so	(private data)
  * Memory mapped usage:
        8192 mmap'd at 0xef7a0000



Program exited normally.
(gdb) quit

Debugger finished
