/*
 *	Copyright 1990 by Rayan S. Zachariassen, all rights reserved.
 *	This will be free software, but only when it is finished.
 */

#include "hostenv.h"
#ifndef	USE_STRCHR

char *
strchr(s, c)
	register char *s;
	register char c;
{
	while (*s && *s != c)
		++s;
	return (*s == '\0' ? 0 : s);
}

char *
strrchr(s, c)
	register char *s;
	register char c;
{
	register char *ss;

	ss = 0;
	while (*s) {
		if (*s == c)
			ss = s;
		++s;
	}
	return ss;
}
#endif	/* !USE_STRCHR */
